@@ -1,18 +1,19 @@
 from __future__ import annotations
 from enum import auto, IntEnum
+import logging
 import numpy as np
 from numpy.typing import ArrayLike
-import os
 import requests
 import time
 import trimesh
-from typing import Callable, Collection, Iterable, Optional
+from typing import Callable, Collection, Iterable, Optional, Union
 from werkzeug.datastructures import ImmutableMultiDict
 
-from .auth import Credentials
-
+from .auth import get_credentials
 from .bracelet_electrodes import generate_bracelet_electrodes as generate_bracelet_electrode_centers
 
+logger = logging.getLogger(__name__)
+
 _API_URL = "http://0.0.0.0:8001"
 
 wait = True
@@ -42,12 +43,8 @@ class SurfaceType(IntEnum):
 DEFAULT_RANDOM_SEED = 0
 
 
-# This object is used to add authentication data to all API requests. By default, authentication credentials can be
-# loaded from a file named ".credentials.txt" stored in the root of the `neurodec` repository, ie, this repository.
-# Such file should contain two lines: the first one being the public key and the second one being the private key.
-# As an alternative, one can create and store here a new Credentials object by passing the public and private keys.
-credentials = Credentials.from_file(os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..",
-                                                 ".credentials.txt"))
+# Default radius to be used for fibers during generation.
+DEFAULT_FIBERS_RADIUS = 2.8e-5
 
 
 class MDTObject:
@@ -84,7 +81,7 @@ class MDTObject:
 
     @classmethod
     def _get_metadata(cls, id: int):
-        response = requests.get(f"{_API_URL}/{cls._url_prefix}/{id}/metadata", json={"credentials": credentials.get()})
+        response = requests.get(f"{_API_URL}/{cls._url_prefix}/{id}/metadata", json={"credentials": get_credentials()})
         if not response.ok:
             raise ValueError("The metadata could not be retrieved.")
 
@@ -94,7 +91,7 @@ class MDTObject:
     def _new_request(cls, json=None):
         if json is None:
             json = {}
-        json.update({"credentials": credentials.get()})
+        json.update({"credentials": get_credentials()})
         response = requests.post(f"{_API_URL}/{cls._url_prefix}", json=json)
         if not response.ok:
             raise ValueError(f"The {cls.__name__} could not be created.")
@@ -129,7 +126,7 @@ class MDTObject:
             self.wait()
 
             response = requests.get(f"{_API_URL}/{self._url_prefix}/{self.id}/data",
-                                    json={"credentials": credentials.get()})
+                                    json={"credentials": get_credentials()})
             if not response.ok:
                 raise RuntimeError("The data of this object could not be retrieved.")
             self._data_json = response.json()
@@ -299,7 +296,7 @@ class Surface(MDTObject):
         surface = cls._new_request(json={"surface": data})
         surface._vertices = vertices
         surface._triangles = triangles
-        surface._label = label if label is not None else f"{surface_type.name.lower}_{surface.id}"
+        surface._label = label if label is not None else f"{surface_type.name.lower()}_{surface.id}"
         return surface
 
     @property
@@ -316,13 +313,39 @@ class Surface(MDTObject):
         return self._label
 
 
+def default_conductivities() -> dict[str, ArrayLike]:
+    """Create a dictionary containing default conductivities for each tissue type.
+
+    Returns:
+        A dictionary in the form {tissue: conductivity} where tissue is one of the strings "skin", "fat", "bone",
+        "muscle" or "electrode" and conductivity is either a scalar (isotropic tissue) or the principal components
+        (anisotropic tissue).
+    """
+    return {
+        "skin": 0.17,
+        "fat": 5.73e-2,
+        "bone": 8.2e-2,
+        "muscle": [0.5, 0.1, 0.1],
+        "electrode": 2.0
+    }
+
+
 class Conductor(MDTObject):
     """A volume conductor"""
 
     _url_prefix: str = "conductor"
 
     @classmethod
-    def new(cls, surfaces: Iterable[Surface], force_computation: bool = False) -> Conductor:
+    def new(
+        cls,
+        surfaces: Iterable[Surface],
+        skin_conductivity: float = 0.17,
+        fat_conductivity: float = 5.73e-2,
+        bone_conductivity: float = 8.2e-2,
+        muscle_conductivity: Optional[Union[float, ArrayLike]] = None,
+        electrode_conductivity: float = 2.0,
+        force_computation: bool = False
+    ) -> Conductor:
         """Create a new volume conductor
 
         Creates a new volume conductor from surfaces. This will initiate a call
@@ -333,6 +356,29 @@ class Conductor(MDTObject):
             surfaces: The surfaces used to define the volume conductor. Must
                 contain at least one OUTER_SKIN surface and one INNER_SKIN
                 surface.
+            skin_conductivity: Conductivity of skin tissue. This tissue is assumed to be isotropic and therefore only
+                one value is required to define the conductivity tensor.
+            fat_conductivity: Conductivity of fat tissue. This tissue is assumed to be isotropic and therefore only one
+                value is required to define the conductivity tensor.
+            bone_conductivity: Conductivity of bone tissue. This tissue is assumed to be isotropic and therefore only
+                one value is required to define the conductivity tensor.
+            muscle_conductivity: Principal conductivity components of muscle tissue. This tissue is assumed to be
+                anisotropic in general, with the principal components being (C1, C2, C2). More precisely, the
+                conductivity C1 is the one associated with the local direction of the fibers. The parameter
+                muscle_conductivity can be:
+                - A scalar Cm. In this case, it is assumed that the conductivity is isotropic, and the vector of
+                  principal components is created as (Cm, Cm, Cm).
+                - A 2-dimensional vector in the form (C1, C2). In this case, a third value is added to complete the
+                  principal components as (C1, C2, C2).
+                - A 3-dimensional vector in the form (C1, C2, C3). The vector is passed as-is to the API. However, a
+                  warning is issued if C2 != C3.
+                - None: the default values (0.5, 0.1, 0.1) are used for the principal components.
+            electrode_conductivity: Conductivity of electrode material. This material is assumed to be isotropic and
+                therefore only one value is required to define the conductivity tensor.
+
+                - A scalar (isotropic tissue).
+                - A vector containing the 3 principal components of the conductivity tensor. This is allowed only for
+                  the MUSCLE tissue type.
             force_computation: Boolean flag that allows to bypass a status check. Normally, the API would check if an
                 object with the requested parameters was already present in the database. If this was the case, the API
                 would not request its computation and would simply return its identifier. By setting this flag to
@@ -345,11 +391,38 @@ class Conductor(MDTObject):
             building a new one.
 
         """
+        # If muscle conductivities are not given, use the default ones. If they are given, make sure they become a list
+        # in the form [c1, c2, c3].
+        if muscle_conductivity is None:
+            muscle_conductivity = [0.5, 0.1, 0.1]
+        elif np.isscalar(muscle_conductivity):
+            muscle_conductivity = [muscle_conductivity] * 3
+        else:
+            muscle_conductivity = np.asarray(muscle_conductivity)
+            if muscle_conductivity.shape == (2,):
+                c1, c2 = muscle_conductivity
+                muscle_conductivity = [c1, c2, c2]
+            elif len(muscle_conductivity) == (3,):
+                c1, c2, c3 = muscle_conductivity
+                if not np.allclose(c2, c3):
+                    logger.warning(f"Received 3 muscle conductivities, with c2 != c3. {c2=}, {c3=}.")
+                muscle_conductivity = [c1, c2, c3]
+            else:
+                raise RuntimeError(f"Principal conductivity components for a muscle must be given as either a scalar or"
+                                   f" an array with 2 or 3 components. An array with shape {muscle_conductivity.shape} "
+                                   f"was received.")
 
         # Encode the surfaces.
         json = {
             "force_computation": force_computation,
             "surfaces": [s.id for s in surfaces],
+            "conductivities": {
+                "skin": skin_conductivity,
+                "fat": fat_conductivity,
+                "bone": bone_conductivity,
+                "muscle": muscle_conductivity,
+                "electrode": electrode_conductivity
+            }
         }
 
         return cls._new_request(json=json)
@@ -518,6 +591,7 @@ class Fibers(MDTObject):
         """
         super().__init__(dictionary)
         self._random_seed = dictionary["random_seed"]
+        self._fibers_3d = None
 
     @property
     def random_seed(self):
@@ -530,11 +604,11 @@ class Fibers(MDTObject):
             plane_origins: ArrayLike,
             plane_normals: ArrayLike,
             n_fibers: Optional[int] = None,
-            fibers_radius: Optional[float] = 2.8e-5,
+            fibers_radius: Optional[float] = None,
             force_computation: bool = False,
             random_seed: int = DEFAULT_RANDOM_SEED
     ) -> Fibers:
-        """Create new fibers for a muscle
+        f"""Create new fibers for a muscle
 
         Creates new fibers from a surface. This will initiate a call
         to the Neurodec Myoelectric Digital Twin (MDT) API that will generate
@@ -549,8 +623,9 @@ class Fibers(MDTObject):
             n_fibers: number of fibers to be generated in the muscle. It cannot
                 be given if fibers_radius is given as well.
             fibers_radius: radius of the fibers. If n_fibers is None, the
-                number of fibers is calculated from this parameter. If n_fibers
-                and this parameter are both None, a default radius is used.
+                number of fibers is calculated from this parameter. If both
+                n_fibers and fibers_radius are None, a default radius of
+                {DEFAULT_FIBERS_RADIUS * 1e6:.1f} micrometers is used.
             force_computation: Boolean flag that allows to bypass a status check. Normally, the API would check if an
                 object with the requested parameters was already present in the database. If this was the case, the API
                 would not request its computation and would simply return its identifier. By setting this flag to
@@ -562,9 +637,7 @@ class Fibers(MDTObject):
             The metadata of the new fibers. If another set of fibers with the
             same parameters already exits, those will be returned instead of
             building new ones.
-
         """
-
         # Convert the planes to numpy arrays and verify the shape.
         plane_origins = np.array(plane_origins, dtype=np.float)
         if plane_origins.shape != (2, 3):
@@ -587,13 +660,31 @@ class Fibers(MDTObject):
             "random_seed": random_seed,
         }
 
+        # Make sure only one parameter is given between n_fibers and fibers_radius.
+        if n_fibers is not None and fibers_radius is not None:
+            raise ValueError(f"Only one argument between 'n_fibers' and 'fibers_radius' can be different from None. "
+                             f"Got {n_fibers=} and {fibers_radius=} instead.")
+
+        # Add to the JSON the number of fibers OR the fiber radius.
         if n_fibers is not None:
             json["n_fibers"] = n_fibers
-        if fibers_radius is not None:
-            json["fibers_radius"] = fibers_radius
+        else:
+            json["fibers_radius"] = fibers_radius if fibers_radius is not None else DEFAULT_FIBERS_RADIUS
 
         return cls._new_request(json=json)
 
+    @property
+    def vertices(self):
+        """Retrieve the fibers generated inside the muscle.
+
+        Return:
+            A NumPy array with shape (n_fibers, n_sections, 3) containing the coordinates of the fiber nodes.
+        """
+        if self._fibers_3d is None:
+            self._retrieve_data()
+            self._fibers_3d = _decode_array(self._data_json, "fibers_3d")
+        return self._fibers_3d
+
 
 class FiberProperties(MDTObject):
     """Properties of individual fibers"""
@@ -969,7 +1060,7 @@ class Electromyography(MDTObject):
             self.wait()
 
             response = requests.get(f"{_API_URL}/{self._url_prefix}/{self.id}/data",
-                                    json={"credentials": credentials.get()})
+                                    json={"credentials": get_credentials()})
             if not response.ok:
                 raise ValueError("The EMG data could not be retrieved.")
             self._data = np.array(response.json()["data"], dtype=np.float)
